/* mlm_udpdelay.c
 *
 * A UDP Packet Delay Plugin form CallCommander
 *
 *   (C) Copyright 2009 Fred Gleason <fredg@paravelsystems.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public
 *   License along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * This is a sample CallCommander Logic Module.  All it does is print
 * a message to stdout for each callback.  (To see the messages, be sure 
 * to specify a Log Path in MDbAdmin->ManageLogicModules).
 *
 * To compile this module, just do:
 * 
 *   gcc -shared -o mlm_test.mlm mlm_udpdelay.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <sys/time.h>

#include <mlm/mlm.h>

#define UDPDELAY_MAX_PACKET_SIZE 1500
#define UDPDELAY_POLL_INTERVAL 100
struct pkt_data {
  char pkt_data[UDPDELAY_MAX_PACKET_SIZE];
  int pkt_len;
  double pkt_timestamp;
};

int udp_handle=-1;
int dest_destinations=0;
char *dest_hostnames=NULL;
uint16_t *dest_ports=NULL;
double delay_secs=0.0;
struct pkt_data *pkt_data=NULL;
unsigned pkt_count=0;


double udpdelay_Timestamp()
{
  struct timeval tv;

  memset(&tv,0,sizeof(tv));
  gettimeofday(&tv,NULL);
  return ((double)tv.tv_sec)+((double)tv.tv_usec/1000000.0);
}


void udpdelay_PushPacket(const char *data,int len)
{
  int i;

  pkt_data=realloc(pkt_data,++pkt_count*sizeof(struct pkt_data));
  for(i=0;i<len;i++) {
    pkt_data[pkt_count-1].pkt_data[i]=data[i];
  }
  pkt_data[pkt_count-1].pkt_len=len;
  pkt_data[pkt_count-1].pkt_timestamp=udpdelay_Timestamp();
}


void udpdelay_PopPacket()
{
  int i;

  pkt_count--;
  for(i=0;i<pkt_count;i++) {
    pkt_data[i]=pkt_data[i+1];
  }
  pkt_data=realloc(pkt_data,sizeof(struct pkt_data)*pkt_count);
}


void udpdelay_SendPacket(const char *data,int len)
{
  int i;

  for(i=0;i<dest_destinations;i++) {
    MLMSendUdp(dest_hostnames+i*PATH_MAX,dest_ports[i],data,len);
  }
}


void MLMStartup(const char *args)
{
  char hostname[PATH_MAX];
  char label[255];

  if((udp_handle=
      MLMOpenUdp(MLMGetIntegerValue(args,"Delay","InputPort",0)))<0) {
    syslog(LOG_ERR,"unable to bind input port %d, exiting...\n",
	   MLMOpenUdp(MLMGetIntegerValue(args,"Delay","InputPort",0)));
  }

  sprintf(label,"DestinationHostname%d",dest_destinations+1);
  strncpy(hostname,MLMGetStringValue(args,"Delay",label,""),PATH_MAX);
  if(strlen(hostname)==0) {
    syslog(LOG_WARNING,"mlm_udpdelay: no destination hostnames specified");
    return;
  }
  while(strlen(hostname)>0) {
    dest_hostnames=realloc(dest_hostnames,(dest_destinations+1)*PATH_MAX);
    strcpy(dest_hostnames+PATH_MAX*dest_destinations,hostname);
    dest_ports=realloc(dest_ports,(dest_destinations+1)*sizeof(uint16_t));
    sprintf(label,"DestinationPort%d",dest_destinations+1);
    dest_ports[dest_destinations]=MLMGetIntegerValue(args,"Delay",label,0);

    dest_destinations++;
    sprintf(label,"DestinationHostname%d",dest_destinations+1);
    strncpy(hostname,MLMGetStringValue(args,"Delay",label,""),PATH_MAX);
  }

  MLMStartTimer(0,UDPDELAY_POLL_INTERVAL,MLM_TIMER_REPEATING);
}


void MLMShutdown()
{
  MLMStopTimer(0);
  MLMCloseUdp(udp_handle);
  free(pkt_data);
  free(dest_hostnames);
  free(dest_ports);
}


void MLMDelayOffsetChanged(int msecs)
{
  delay_secs=(double)msecs/1000.0;
}


void MLMUdpReceived(int handle,const char *data,int len)
{
  if(delay_secs==0.0) {
    udpdelay_SendPacket(data,len);
  }
  else {
    udpdelay_PushPacket(data,len);
  }
}


void MLMTimerExpired(int timernum)
{
  double current_time;

  current_time=udpdelay_Timestamp();

  while(pkt_count>0) {
    if((pkt_data[0].pkt_timestamp+delay_secs)>current_time) {
      return;
    }
    udpdelay_SendPacket(pkt_data[0].pkt_data,pkt_data[0].pkt_len);
    udpdelay_PopPacket();
  }
}
